import abc
import pandas as pd
import pygit2 as git

from tqdm import tqdm
from tqdm.contrib.concurrent import thread_map

from tools.timeit import Timeit


def map_signature(mailmap, signature: git.Signature):
    # the unmapped email is used on purpose
    email = signature.email
    try:
        mapped_signature = mailmap.resolve_signature(signature)
        name = mapped_signature.name
    except ValueError:
        name = signature.name
        if not name:
            name = "Empty Empty"
            # warnings.warn(f"{str(e)}. Name will be replaced with '{name}'")
        if not email:
            email = "empty@empty.empty"
            # warnings.warn(f"{str(e)}. Email will be replaced with '{email}'")
    return name, email


class History(abc.ABC):

    def __init__(self, repository: git.Repository, branch: str = "master"):
        self.repo = repository
        self.branch = branch
        self.cache = None
        self.mailmap = git.Mailmap.from_repository(self.repo)

    def as_dataframe(self):
        data = self.fetch()
        df = pd.DataFrame(data)
        return self._optimize(df)

    @abc.abstractmethod
    def fetch(self):
        pass

    @abc.abstractmethod
    def _optimize(self, df: pd.DataFrame):
        return df

    @property
    def commits_walker(self):
        return self.repo.walk(self.repo.head.target, git.GIT_SORT_TOPOLOGICAL)

    def get_commits_count(self):
        return sum(1 for _ in self.commits_walker)


class WholeHistory(History):

    @Timeit("Fetching whole history data")
    def fetch(self):
        records = []
        for commit in tqdm(self.commits_walker, total=self.get_commits_count()):
            author_name, author_email = map_signature(self.mailmap, commit.author)

            is_merge_commit = False
            insertions, deletions = 0, 0
            if len(commit.parents) == 0:  # initial commit
                st = commit.tree.diff_to_tree(swap=True).stats
                insertions, deletions = st.insertions, st.deletions
            elif len(commit.parents) == 1:
                parent_commit = commit.parents[0]
                st = self.repo.diff(parent_commit, commit).stats
                insertions, deletions = st.insertions, st.deletions
            # case len(commit.parents) > 1 corresponds to a merge commit
            # merge commits are ignored: changes in merge commits are normally because of integration issues
            else:
                is_merge_commit = True

            records.append({'commit_sha': commit.hex[:7],
                            'is_merge_commit': is_merge_commit,
                            'author_name': author_name,
                            'author_email': author_email,
                            'author_tz_offset': commit.author.offset,
                            'author_timestamp': commit.author.time,
                            'review_duration': commit.committer.time - commit.author.time,
                            'insertions': insertions,
                            'deletions': deletions})
        return records

    def _optimize(self, df: pd.DataFrame):
        df['author_name'] = pd.Categorical(df['author_name'])
        df['author_email'] = pd.Categorical(df['author_email'])
        return df


class LinearHistory(History):

    @Timeit("Fetching linear history data")
    def fetch(self):
        records = []
        for commit in tqdm(self.commits_walker, total=self.get_commits_count()):

            insertions, deletions = 0, 0
            if len(commit.parents) == 0:  # initial commit
                st = commit.tree.diff_to_tree(swap=True).stats
                insertions, deletions = st.insertions, st.deletions
            elif len(commit.parents) >= 1:
                parent_commit = commit.parents[0]
                st = self.repo.diff(parent_commit, commit).stats
                insertions, deletions = st.insertions, st.deletions

            records.append({'commit_sha': commit.hex[:7],
                            'committer_timestamp': commit.committer.time,
                            'files_count': len(commit.tree.diff_to_tree()),
                            'insertions': insertions,
                            'deletions': deletions})
        return records

    def _optimize(self, df: pd.DataFrame):
        return super()._optimize(df)

    @property
    def commits_walker(self):
        walker = super(LinearHistory, self).commits_walker
        walker.simplify_first_parent()
        return walker


class BlameData:
    """
    Class to fetch raw data about repository state at certain revision
    """
    def __init__(self, repository: git.Repository, revision: str = None):
        self.repo = repository
        self.mailmap = git.Mailmap.from_repository(self.repo)
        self.revision_commit = self.repo.revparse_single(revision) if revision else self.repo.head.peel()

    def _get_data_from_blame_hunk(self, blame_hunk):
        hunk_committer = blame_hunk.final_committer
        if not hunk_committer:
            # if committer configured an empty email when created commit
            # blame hunk corresponding to that commit will produce a None signature
            # the following substitutes hunk's final committer with an author of the commit
            hunk_committer = self.repo[blame_hunk.orig_commit_id].author
        committer_name, _ = map_signature(self.mailmap, hunk_committer)
        return [committer_name, blame_hunk.lines_in_hunk, hunk_committer.time]

    def blame_file(self, file_path):
        blob_blame = self.repo.blame(file_path)
        blame_info = [self._get_data_from_blame_hunk(blame_hunk) + [file_path] for blame_hunk in blob_blame]
        return blame_info

    @Timeit("Fetching blame data")
    def fetch(self):
        submodules_paths = self.repo.listall_submodules()
        diff_to_tree = self.revision_commit.tree.diff_to_tree()
        files_to_blame = [p.delta.new_file.path for p in diff_to_tree
                          if not p.delta.is_binary and p.delta.new_file.path not in submodules_paths]

        results = thread_map(self.blame_file, files_to_blame)
        return [rec for val in results for rec in val]

    def as_dataframe(self):
        data = self.fetch()
        df = pd.DataFrame(data, columns=["committer_name", "lines_count", "timestamp", "filepath"])
        # this saves some memory
        df["committer_name"] = pd.Categorical(df["committer_name"])
        df["filepath"] = pd.Categorical(df["filepath"])
        return df


class FilesData:
    """
    Class to fetch raw data about repository state at certain revision
    """
    def __init__(self, repository: git.Repository, revision: str = None):
        self.repo = repository
        self.revision_commit = self.repo.revparse_single(revision) if revision else self.repo.head.peel()

    @Timeit("Fetching files data")
    def _fetch(self):
        submodules_paths = self.repo.listall_submodules()
        head_commit_tree = self.revision_commit.tree.diff_to_tree(swap=True)
        records = []
        for p in head_commit_tree:
            filepath = p.delta.new_file.path
            if filepath not in submodules_paths:
                records.append({
                    "file": filepath,
                    "is_binary": p.delta.is_binary,
                    "size_bytes": p.delta.new_file.size,
                    "lines_count": p.line_stats[1]
                })
        return records

    def as_dataframe(self):
        data = self._fetch()
        df = pd.DataFrame(data)
        return df


class TagsData:
    def __init__(self, repository: git.Repository):
        """
        :param path: path to a repository
        """
        self.repo = repository
        self.mailmap = git.Mailmap.from_repository(self.repo)

    @Timeit("Fetching tags info")
    def fetch(self):
        # TODO: this should perhaps be a part of WholeHistory
        tag_refs = {refobj.peel().oid: refobj
                    for refobj in self.repo.listall_reference_objects() if refobj.name.startswith('refs/tags')}

        result = []
        tag_ref = None
        is_symbolic_reference = False
        for commit in self.repo.walk(self.repo.head.target, git.GIT_SORT_TOPOLOGICAL):
            author_name, _ = map_signature(self.mailmap, commit.author)
            if commit.oid in tag_refs:
                tag_ref: git.Reference = tag_refs[commit.oid]
                is_symbolic_reference = tag_ref.target.hex == commit.hex

            if tag_ref is not None:
                if not is_symbolic_reference:
                    tag = self.repo[tag_ref.target]
                    tagger_name, _ = map_signature(self.mailmap, tag.tagger)
                    tag_metadata = {
                        "tag_name": tag.name,
                        "tagger_name": tagger_name,
                        "tagger_time": tag.tagger.time,
                    }
                else:
                    tag_metadata = {
                        "tag_name": tag_ref.shorthand,
                        "tagger_name": None,
                        "tagger_time": -1,
                    }
            else:
                tag_metadata = {
                    "tag_name": None,
                    "tagger_name": None,
                    "tagger_time": -1,
                }
            tag_metadata["commit_author"] = author_name
            tag_metadata["commit_time"] = commit.author.time
            tag_metadata["is_merge"] = len(commit.parents) > 1
            result.append(tag_metadata)

        return result

    def as_dataframe(self):
        raw_data = pd.DataFrame(self.fetch())
        # TODO: optimize memory by changing some datatypes
        # .astype('bool')
        return raw_data



import os
import datetime
import calendar
import collections
import json
from jinja2 import Environment, FileSystemLoader
import pandas as pd

from analysis.gitrepository import GitRepository
from tools.configuration import Configuration
from tools import packages_info

from . import colormaps
from .html_page import HtmlPage, JsPlot

HERE = os.path.dirname(os.path.abspath(__file__))


class HTMLReportCreator:
    recent_activity_period_weeks = 32
    assets_subdir = "assets"
    templates_subdir = "templates"

    def __init__(self, config: Configuration, repository: GitRepository):
        self.path = None
        self.configuration = config
        self.assets_path = os.path.join(HERE, self.assets_subdir)
        self.git_repository_statistics = repository
        self.has_tags_page = config.do_process_tags()
        self._time_sampling_interval = "W"
        self._do_generate_index_page = False
        self._is_blame_data_allowed = False
        self._max_orphaned_extensions_count = 0

        templates_dir = os.path.join(HERE, self.templates_subdir)
        self.j2_env = Environment(loader=FileSystemLoader(templates_dir), trim_blocks=True)
        self.j2_env.filters['to_month_name_abr'] = lambda im: calendar.month_abbr[im]
        self.j2_env.filters['to_weekday_name'] = lambda i: calendar.day_name[i]
        self.j2_env.filters['to_ratio'] = lambda val, max_val: (float(val) / max_val) if max_val != 0 else 0
        self.j2_env.filters['to_percentage'] = lambda val, max_val: (100 * float(val) / max_val) if max_val != 0 else 0
        colors = colormaps.colormaps[self.configuration['colormap']]
        self.j2_env.filters['to_heatmap'] = lambda val, max_val: "%d, %d, %d" % colors[int(float(val) / max_val * (len(colors) - 1))]

    def set_time_sampling(self, offset: str):
        """
        :param offset: any valid string composed of Pandas' offset aliases
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
        """
        self._time_sampling_interval = offset
        return self

    def allow_blame_data(self):
        self._is_blame_data_allowed = True

    def generate_index_page(self, do_generate: bool = True):
        self._do_generate_index_page = do_generate
        return self

    def set_max_orphaned_extensions_count(self, count):
        self._max_orphaned_extensions_count = count
        return self

    def _clamp_orphaned_extensions(self, extensions_df: pd.DataFrame, group_name: str = "~others~"):
        # Group together all extensions used only once (probably not really extensions)
        is_orphan = extensions_df["files_count"] <= self._max_orphaned_extensions_count
        excluded = extensions_df[is_orphan]
        print(excluded.shape)
        if excluded.shape[0] > 0:
            excluded_summary = excluded.agg({"size_bytes": ["sum"], "lines_count": ["sum", "count"]})
            orphans_summary_df = pd.DataFrame(data=[{
                "files_count": excluded_summary['lines_count']['count'],
                "lines_count": excluded_summary['lines_count']['sum'],
                "size_bytes": excluded_summary['size_bytes']['sum'].astype('int32')
            }], index=[(False, group_name)]) # index is a tuple (is_binary, extension)

            extensions_df = extensions_df[~is_orphan].sort_values(by="files_count", ascending=False)
            # and we do not sort after we appended "orphans", as we want them to appear at the end
            extensions_df = extensions_df.append(orphans_summary_df, sort=False)

        return extensions_df

    def _get_recent_activity_data(self):
        recent_weekly_commits = self.git_repository_statistics.\
            get_recent_weekly_activity(self.recent_activity_period_weeks)

        values = [
            {
                'x': int(self.recent_activity_period_weeks - i - 1),
                'y': int(commits)
            } for i, commits in enumerate(recent_weekly_commits)
        ]
        graph_data = {
            "config": {
                "noData": "No recent activity.",
                "padData": True,
                "showXAxis": True,
                "xDomain": [self.recent_activity_period_weeks - 1, 0]
            },
            "data": [
                {'key': "Commits", 'color': "#9400D3", 'values': values}
            ]
        }

        return graph_data

    def _bundle_assets(self):
        from distutils.dir_util import copy_tree
        # copy assets to report output folder
        assets_local_abs_path = os.path.join(self.path, self.assets_subdir)
        copy_tree(src=self.assets_path, dst=assets_local_abs_path)
        # relative path to assets to embed into html pages
        self.assets_path = os.path.relpath(assets_local_abs_path, self.path)

    def _squash_authors_history(self, authors_history, max_authors_count):
        authors = self.git_repository_statistics.authors.sort(by='commits_count').names()

        most_productive_authors = authors[:max_authors_count]
        rest_authors = authors[max_authors_count:]

        most_productive_authors_history = authors_history[most_productive_authors] \
            .asfreq(freq=self._time_sampling_interval, fill_value=0) \
            .cumsum()
        rest_authors_history = authors_history[rest_authors].sum(axis=1)\
            .asfreq(freq=self._time_sampling_interval, fill_value=0)\
            .cumsum()

        most_productive_authors_history.columns = most_productive_authors_history.columns.add_categories(['Others'])
        most_productive_authors_history['Others'] = rest_authors_history.values
        return most_productive_authors_history

    def create(self, path):
        self.path = path

        if self.configuration.is_report_relocatable():
            self._bundle_assets()
        HtmlPage.set_assets_path(self.assets_path)

        pages = [
            self.make_general_page(),
            self.make_activity_page(),
            self.make_authors_page(),
            self.make_files_page(),
        ]
        if self.has_tags_page:
            pages.append(self.make_tags_page())
        pages.append(self.make_about_page())

        # render and save all pages
        for page in pages:
            rendered_page = page.render(self.j2_env, linked_pages=pages)
            page.save(self.path, rendered_page)

        if self._do_generate_index_page:
            # make the landing page for a web server
            from shutil import copyfile
            copyfile(os.path.join(path, "general.html"), os.path.join(path, "index.html"))

    def make_general_page(self):
        first_commit_datetime = datetime.datetime.fromtimestamp(self.git_repository_statistics.first_commit_timestamp)
        last_commit_datetime = datetime.datetime.fromtimestamp(self.git_repository_statistics.last_commit_timestamp)

        project_data = {
            "name": self.git_repository_statistics.name,
            "branch": self.git_repository_statistics.branch,
            "age": (last_commit_datetime - first_commit_datetime).days,
            "active_days_count": self.git_repository_statistics.active_days_count,
            "commits_count": self.git_repository_statistics.total_commits_count,
            "merge_commits_count": self.git_repository_statistics.merge_commits_count,
            "authors_count": self.git_repository_statistics.authors.count(),
            "files_count": self.git_repository_statistics.head.files_count,
            "total_lines_count": self.git_repository_statistics.total_lines_count,
            "added_lines_count": self.git_repository_statistics.total_lines_added,
            "removed_lines_count": self.git_repository_statistics.total_lines_removed,
            "first_commit_date": first_commit_datetime,
            "last_commit_date": last_commit_datetime,
        }

        generation_data = {
            "datetime": datetime.datetime.today().strftime('%Y-%m-%d %H:%M')
        }

        page = HtmlPage(name="General",
                        project=project_data,
                        generation=generation_data)
        return page

    def make_activity_page(self):
        # TODO: this conversion from old 'data' to new 'project data' should perhaps be removed in future
        project_data = {
            'timezones_activity': collections.OrderedDict(
                sorted(self.git_repository_statistics.timezones_distribution.items(), key=lambda n: int(n[0]))),
            'month_in_year_activity': self.git_repository_statistics.month_of_year_distribution.to_dict()
        }

        wd_h_distribution = self.git_repository_statistics.weekday_hour_distribution.astype('int32')
        project_data['weekday_hourly_activity'] = wd_h_distribution
        project_data['weekday_hour_max_commits_count'] = wd_h_distribution.max().max()
        project_data['weekday_activity'] = wd_h_distribution.sum(axis=1)
        project_data['hourly_activity'] = wd_h_distribution.sum(axis=0)

        page = HtmlPage(name='Activity', project=project_data)
        page.add_plot(self.make_activity_plot())

        return page

    def make_activity_plot(self) -> JsPlot:
        recent_activity = self._get_recent_activity_data()

        # Commits by current year's months
        current_year = datetime.date.today().year
        current_year_monthly_activity = self.git_repository_statistics.history('m')
        current_year_monthly_activity = current_year_monthly_activity \
            .loc[current_year_monthly_activity.index.year == current_year]

        current_year_monthly_activity = pd.Series(current_year_monthly_activity.values,
                                                  index=current_year_monthly_activity.index.month).to_dict()
        values = [
            {
                'x': imonth,
                'y': current_year_monthly_activity.get((imonth + 1), 0)
            } for imonth in range(0, 12)
        ]
        by_month = {
            "yAxis": {"axisLabel": "Commits in %d" % current_year},
            "xAxis": {"rotateLabels": -90, "ticks": len(values)},
            "data": [
                {"key": "Commits", "color": "#9400D3", "values": values}
            ]
        }

        # Commits by year
        yearly_activity = self.git_repository_statistics.history('Y')
        values = [{'x': int(x.year), 'y': int(y)} for x, y in zip(yearly_activity.index, yearly_activity.values)]

        by_year = {
            "xAxis": {"rotateLabels": -90, "ticks": len(values)},
            "yAxis": {"axisLabel": "Commits"},
            "data": [
                {"key": "Commits", "color": "#9400D3", "values": values}
            ]
        }

        review_duration = [
            {
                "label": label,
                "value": count
            }
            for label, count in self.git_repository_statistics.review_duration_distribution.items()
        ]

        activity_plot = JsPlot('activity.js',
                               commits_by_month=json.dumps(by_month),
                               commits_by_year=json.dumps(by_year),
                               recent_activity=json.dumps(recent_activity),
                               review_duration=json.dumps(review_duration),
                               )
        return activity_plot

    def make_authors_page(self):
        authors_summary = self.git_repository_statistics.authors.summary \
            .sort_values(by="commits_count", ascending=False)

        top_authors_statistics = authors_summary[:self.configuration['max_authors']]
        non_top_authors_names = authors_summary[self.configuration['max_authors']:]['author_name'].values
        project_data = {
            'top_authors_statistics': top_authors_statistics,
            'non_top_authors': non_top_authors_names,
            'authors_top': self.configuration['authors_top'],
            'total_commits_count': self.git_repository_statistics.total_commits_count,
            'total_lines_count': self.git_repository_statistics.total_lines_count,
            'is_blame_data_available': self._is_blame_data_allowed,
        }

        if self._is_blame_data_allowed:
            project_data.update({
                'top_knowledge_carriers': self.git_repository_statistics.head.get_top_knowledge_carriers()
                    .head(self.configuration['authors_top'])
            })

        raw_authors_data = self.git_repository_statistics.get_authors_ranking_by_month()
        ordered_months = raw_authors_data.index.get_level_values(0).unique().sort_values(ascending=False)
        project_data['months'] = []
        for yymm in ordered_months[0:self.configuration['max_authors_of_months']]:
            authors_in_month = raw_authors_data.loc[yymm]
            project_data['months'].append({
                'date': yymm,
                'top_author': {'name': authors_in_month.index[0], 'commits_count': authors_in_month.iloc[0]},
                'next_top_authors': ', '.join(list(authors_in_month.index[1:5])),
                'all_commits_count': authors_in_month.sum(),
                'total_authors_count': authors_in_month.size
            })

        project_data['years'] = []
        raw_authors_data = self.git_repository_statistics.get_authors_ranking_by_year()
        max_top_authors_index = self.configuration['authors_top'] + 1
        ordered_years = raw_authors_data.index.get_level_values(0).unique().sort_values(ascending=False)
        for y in ordered_years:
            authors_in_year = raw_authors_data.loc[y]
            project_data['years'].append({
                'date': y,
                'top_author': {'name': authors_in_year.index[0], 'commits_count': authors_in_year.iloc[0]},
                'next_top_authors': ', '.join(list(authors_in_year.index[1:max_top_authors_index])),
                'all_commits_count': authors_in_year.sum(),
                'total_authors_count': authors_in_year.size,
            })

        page = HtmlPage('Authors', project=project_data)
        page.add_plot(self.make_authors_plot())
        return page

    def make_authors_plot(self) -> JsPlot:
        max_authors_per_plot_count = self.configuration['max_plot_authors_count']

        authors_activity_history = self.git_repository_statistics.authors.history(self._time_sampling_interval)

        authors_commits_history = self._squash_authors_history(authors_activity_history.commits_count,
                                                               max_authors_per_plot_count)
        authors_added_lines_history = self._squash_authors_history(authors_activity_history.insertions,
                                                                   max_authors_per_plot_count)

        # "Added lines" graph
        data = []
        for author in authors_added_lines_history:
            authorstats = {'key': author}
            series = authors_added_lines_history[author]
            authorstats['values'] = [{'x': int(x.timestamp()) * 1000, 'y': int(y)} for x, y in zip(series.index, series.values)]
            data.append(authorstats)

        lines_by_authors = {
            "data": data
        }

        # "Commit count" and streamgraph
        # TODO move the "added lines" into the same JSON to save space and download time
        data = []
        for author in authors_commits_history:
            authorstats = {'key': author}
            series = authors_commits_history[author]
            stream = series.diff().fillna(0)
            authorstats['values'] = [[int(x.timestamp() * 1000), int(y), int(z)] for x, y, z in zip(series.index, series.values, stream.values)]
            data.append(authorstats)

        commits_by_authors = {
            "data": data
        }

        email_domains_distribution = self.git_repository_statistics.domains_distribution\
            .sort_values(ascending=False)
        if self.configuration['max_domains'] < email_domains_distribution.shape[0]:
            top_domains = email_domains_distribution[:self.configuration['max_domains']]
            other_domains = email_domains_distribution[self.configuration['max_domains']:].sum()
            email_domains_distribution = top_domains.append(pd.Series(other_domains, index=["Others"]))

        from collections import OrderedDict
        email_domains_distribution = email_domains_distribution.to_dict(OrderedDict)

        # Domains
        domains = {
            "data": [{"key": domain, "y": commits_count} for domain, commits_count in email_domains_distribution.items()]
        }

        if self._is_blame_data_allowed:
            # sort by contribution
            sorted_contribution = self.git_repository_statistics.head.authors_contribution.sort_values(ascending=False)
            # limit to only top authors
            if sorted_contribution.shape[0] > max_authors_per_plot_count + 1:
                rest_contributions = sorted_contribution[max_authors_per_plot_count:].sum()
                sorted_contribution = sorted_contribution[:max_authors_per_plot_count]
                # at this point index is a CategoricalIndex and without next line cannot accept new category: "others"
                sorted_contribution.index = sorted_contribution.index.to_list()
                sorted_contribution = sorted_contribution.append(pd.Series(rest_contributions, index=["others"]))
            sorted_contribution = sorted_contribution.to_dict(OrderedDict)
            # Contribution plot data
            contribution = {
                "data": [{"key": name, "y": lines_count} for name, lines_count in sorted_contribution.items()]
            }
        else:
            contribution = {}

        authors_plot = JsPlot('authors.js',
                              lines_by_authors=json.dumps(lines_by_authors),
                              commits_by_authors=json.dumps(commits_by_authors),
                              domains=json.dumps(domains),
                              contribution=json.dumps(contribution)
                              )
        return authors_plot

    def make_files_page(self):
        file_ext_summary = self.git_repository_statistics.head.files_extensions_summary
        if self._max_orphaned_extensions_count > 0:
            file_ext_summary = self._clamp_orphaned_extensions(file_ext_summary)
        else:
            file_ext_summary = file_ext_summary.sort_values(by="files_count", ascending=False)

        project_data = {
            'total_files_count': self.git_repository_statistics.head.files_count,
            'total_lines_count': self.git_repository_statistics.total_lines_count,
            'size': self.git_repository_statistics.head.size,
            'file_summary': file_ext_summary,
            'is_blame_data_available': self._is_blame_data_allowed
        }
        if self._is_blame_data_allowed:
            project_data.update({
                'top_files_by_contributors_count': self.git_repository_statistics.head.get_top_files_by_contributors_count(),
                'monoauthor_files_count': self.git_repository_statistics.head.monoauthor_files.count(),
                'lost_knowledge_ratio': self.git_repository_statistics.head.get_lost_knowledge_percentage()
            })

        page = HtmlPage('Files', project=project_data)
        page.add_plot(self.make_files_plot())
        return page

    def make_files_plot(self) -> JsPlot:
        hst = self.git_repository_statistics.linear_history(self._time_sampling_interval).copy()
        hst["epoch"] = (hst.index - pd.Timestamp("1970-01-01 00:00:00+00:00")) // pd.Timedelta('1s') * 1000

        files_count_ts = hst[["epoch", 'files_count']].rename(columns={"epoch": "x", 'files_count': "y"})\
            .to_dict('records')
        lines_count_ts = hst[["epoch", 'lines_count']].rename(columns={"epoch": "x", 'lines_count': "y"})\
            .to_dict('records')
        graph_data = {
            "data": [
                {"key": "Files", "color": "#9400d3", "type": "line", "yAxis": 1, "values": files_count_ts},
                {"key": "Lines", "color": "#d30094", "type": "line", "yAxis": 2, "values": lines_count_ts},
            ]
        }

        files_plot = JsPlot('files.js', json_data=json.dumps(graph_data))
        return files_plot

    def make_tags_page(self):
        if 'max_recent_tags' not in self.configuration:
            tags = list(self.git_repository_statistics.tags.all())
        else:
            tags = [next(self.git_repository_statistics.tags) for _ in range(self.configuration['max_recent_tags'])]

        project_data = {
            'tags': tags,
            # this is total tags count, generally len(tags) != total_tags_count
            'tags_count': self.git_repository_statistics.tags.count
        }

        page = HtmlPage(name='Tags', project=project_data)
        return page

    def make_about_page(self):
        repostat_version = self.configuration.get_release_data_info()['develop_version']
        repostat_version_date = self.configuration.get_release_data_info()['user_version']
        page_data = {
            "version": f"{repostat_version} ({repostat_version_date})",
            "tools": [packages_info.get_pygit2_info(),
                      packages_info.get_jinja_info()],
            "contributors": self.configuration.get_release_data_info()['contributors']
        }

        page = HtmlPage('About', repostat=page_data)
        return page



import pandas as pd
import pygit2 as git
import warnings
from datetime import datetime
import os
import pytz

from tools import split_email_address
from .gitdata import WholeHistory as GitWholeHistory
from .gitdata import LinearHistory as GitLinearHistory
from .gitrevision import GitRevision
from .gitauthors import GitAuthors
from .gittags import GitTags


class GitRepository:
    def __init__(self, path: str):
        """
        :param path: path to a repository
        """
        self.repo = git.Repository(path)
        self.branch = self.repo.head.shorthand
        self.whole_history_df = GitWholeHistory(self.repo).as_dataframe()
        self.linear_history_df = GitLinearHistory(self.repo).as_dataframe()
        self._head_revision = None
        self._tags = None
        self._name = None

    @property
    def name(self):
        if self._name is None:
            # remove trailing slash
            head, _ = os.path.split(self.repo.path)
            # remove '.git' subfolder
            head, _ = os.path.split(head)
            # get the folder name containing '.git'
            _, self._name = os.path.split(head)
        return self._name

    @property
    def head(self):
        if not self._head_revision:
            self._head_revision = GitRevision(self.repo, 'HEAD')
        return self._head_revision

    @property
    def tags(self):
        if not self._tags:
            self._tags = GitTags(self.repo)
        return self._tags

    @property
    def total_commits_count(self):
        return self.whole_history_df.shape[0]

    @property
    def merge_commits_count(self):
        return self.whole_history_df['is_merge_commit'].sum()

    @property
    def total_lines_added(self):
        return self.linear_history_df['insertions'].sum()

    @property
    def total_lines_removed(self):
        return self.linear_history_df['deletions'].sum()

    @property
    def total_lines_count(self):
        return self.total_lines_added - self.total_lines_removed

    @property
    def first_commit_timestamp(self):
        return self.whole_history_df["author_timestamp"].min()

    @property
    def last_commit_timestamp(self):
        return self.whole_history_df["author_timestamp"].max()

    @property
    def active_days_count(self):
        # Note, calculations here are done in UTC, calculation in local tz may give slightly different days count
        count = pd.to_datetime(self.whole_history_df['author_timestamp'], unit='s').\
            dt.strftime('%Y-%m-%d').unique().size
        return count

    @property
    def review_duration_distribution(self):
        duration_bins = [pd.Timedelta('0s').total_seconds(),
                         pd.Timedelta('1s').total_seconds(),
                         pd.Timedelta('1 hours').total_seconds(),
                         pd.Timedelta('1 day').total_seconds(),
                         pd.Timedelta('2 days').total_seconds(),
                         pd.Timedelta('1W').total_seconds(),
                         pd.Timedelta('2W').total_seconds(),
                         pd.Timedelta(30, unit='D').total_seconds(),
                         pd.Timedelta(183, unit='D').total_seconds(),
                         pd.Timedelta(3 * 365, unit='D').total_seconds()]

        review_duration_ts = self.whole_history_df['review_duration']
        review_time_binned = pd.cut(review_duration_ts, bins=duration_bins, include_lowest=True,
                                    labels=['= 0s',
                                            '< 1hour',
                                            '< 1day',
                                            '< 2days',
                                            '< 1week',
                                            '< 2weeks',
                                            '< 1month',
                                            '< 6 months',
                                            '< 3 years'
                                            ])
        return review_time_binned.value_counts().sort_index()

    @property
    def timezones_distribution(self):
        # first group commits by timezones' offset given in minutes
        ts = self.whole_history_df['author_tz_offset'].groupby(self.whole_history_df['author_tz_offset']).count()

        dummy_timestamp = pd.Timestamp(datetime.utcnow().replace(tzinfo=pytz.utc))
        # transform tz offsets (given as index of ts) in minutes to strings formatted as strftime('%z')
        # TODO: move this formatting outside of statistics
        formatted_offsets_ts = ts.reset_index(name="counts")['author_tz_offset'].apply(
            # in order to use strftime('%z') formatter, one needs to have a valid timezone aware pd.Timestamp
            # the actual date is not relevant here, so dummy timestamp is used
            lambda x: dummy_timestamp.tz_convert(pytz.FixedOffset(x)).strftime('%z'))

        # re-create series with formatted index
        return pd.Series(ts.values, index=formatted_offsets_ts.values).to_dict()

    @staticmethod
    def _fetch_domain_from_email(email):
        try:
            _, domain = split_email_address(email)
        except ValueError as ex:
            warnings.warn(str(ex))
            domain = "unknown"
        return domain

    @property
    def domains_distribution(self):
        domains_ts = self.whole_history_df['author_email'].apply(self._fetch_domain_from_email)
        return domains_ts.groupby(by=domains_ts.values).count()

    def get_recent_weekly_activity(self, recent_weeks_count: int):
        """
        Calculates contributors' weekly activity (number of commits per week)
        :param recent_weeks_count: time period in weeks
        :return: sampled number of commits
        """
        assert recent_weeks_count > 0

        today = pd.Timestamp.today()
        if today.weekday() == 6:  # 'today' is Sunday
            # set Sunday within a week as last day of recent activity ('today' is already in next week)
            last_activity_date = today + pd.Timedelta(weeks=1)
        else:
            # set last day of recent activity interval as next Monday
            last_activity_date = today + pd.Timedelta(days=-today.weekday(), weeks=1)
        # Monday `recent_weeks_count` weeks ago
        start_activity_date = last_activity_date - pd.Timedelta(weeks=recent_weeks_count)

        # TODO: committer timestamp better reflects recent activity on a current branch
        ts = pd.to_datetime(self.whole_history_df['author_timestamp'], unit='s')
        ddf = pd.DataFrame({'timestamp': ts[ts >= start_activity_date]})
        # weekly intervals
        intervals = pd.date_range(end=last_activity_date, periods=recent_weeks_count + 1, freq='W-SUN', normalize=True)
        # sample commits number by weekly intervals
        histogram = pd.cut(ddf.timestamp, bins=intervals)

        result_ts = histogram.groupby(histogram.values).count()
        return result_ts.values

    def get_authors_ranking_by_year(self):
        """
        Top authors by all years of repo existence as pandas timeseries, e.g
        timestamp  author_name
        2007       Author3        1
        2020       Author1        2
                   Author2        1

        :return: Pandas multiindex timeseries:  (<year>, <author_name>) -> <commits count>
        """
        df = pd.DataFrame({'author_name': self.whole_history_df['author_name'],
                           'timestamp': pd.to_datetime(self.whole_history_df['author_timestamp'], unit='s')})
        ts_agg = df.groupby([df.timestamp.dt.year, df.author_name]).size()
        # https://stackoverflow.com/questions/27842613/pandas-groupby-sort-within-groups
        # group by the first level of the index
        ts_agg = ts_agg.groupby(level=0, group_keys=False)
        # then sort each group
        res = ts_agg.apply(lambda x: x.sort_values(ascending=False))

        return res

    def get_authors_ranking_by_month(self):
        """
        Top authors by all month of repo existence as pandas timeseries, e.g
        timestamp  author_name
        2007-07    Author3        2
        2020-02    Author2        1
                   Author1        1

        :return: Pandas multiindex timeseries:  (<year>-<month>, <author_name>) -> <commits count>
        """

        df = pd.DataFrame({'author_name': self.whole_history_df['author_name'],
                           'timestamp': pd.to_datetime(self.whole_history_df['author_timestamp'], unit='s')
                          .dt.strftime('%Y-%m')})

        # https://stackoverflow.com/questions/27842613/pandas-groupby-sort-within-groups
        ts_agg = df.groupby([df.timestamp, df.author_name]).size()\
            .groupby(level=0, group_keys=False)

        # sort each group by value
        return ts_agg.apply(lambda x: x.sort_values(ascending=False))

    @property
    def authors(self) -> GitAuthors:
        if not hasattr(self, '_authors'):
            setattr(self, '_authors', GitAuthors(self.whole_history_df))
        return getattr(self, '_authors')

    @property
    def month_of_year_distribution(self):
        ts = pd.to_datetime(self.whole_history_df['author_timestamp'], unit='s', utc=True)
        return ts.groupby(ts.dt.month).count()

    @property
    def weekday_hour_distribution(self):
        df = self.whole_history_df[['author_timestamp']].copy()
        # Weekday activity should be calculated in local timezones
        # https://stackoverflow.com/questions/36648995/how-to-add-timezone-offset-to-pandas-datetime
        df['datetime'] = pd.to_datetime(self.whole_history_df['author_timestamp'], unit='s', utc=True) + \
                         pd.TimedeltaIndex(self.whole_history_df['author_tz_offset'], unit='m')
        df['weekday'] = df.datetime.dt.weekday
        df['hour'] = df.datetime.dt.hour

        df = df.pivot_table(
            index=df['weekday'],
            columns=df['hour'],
            values='datetime',
            aggfunc='count'
        ).fillna(0)
        return df

    def history(self, sampling):
        # this is "commits history" with timeline defined by "author_timestamp", i.e. by time when a commit was created
        df = self.whole_history_df[['author_timestamp']].copy()
        df['datetime'] = pd.to_datetime(self.whole_history_df['author_timestamp'], unit='s', utc=True)
        df = df.set_index(df['datetime'])\
            .rename(columns={'datetime': 'commits_count'})['commits_count']\
            .groupby(pd.Grouper(freq=sampling)) \
            .count()
        return df

    def linear_history(self, sampling):
        # this is "modifications history" with timeline defined by "committer_timestamp", i.e. by time when a
        # commit was incorporated (via create/amend/rebase) into branch
        df = self.linear_history_df[['committer_timestamp', 'files_count', 'insertions', 'deletions']].copy()
        df['datetime'] = pd.to_datetime(self.linear_history_df['committer_timestamp'], unit='s', utc=True)
        df = df.set_index(df['datetime'])

        wh_grouped = df[['files_count', 'insertions', 'deletions']].groupby(pd.Grouper(freq=sampling))
        result = wh_grouped[['insertions', 'deletions']].sum().cumsum()
        result['files_count'] = wh_grouped['files_count'].mean().fillna(method='ffill')
        result['lines_count'] = result['insertions'] - result['deletions']
        return result



import pygit2
import jinja2


def get_pygit2_info():
    return f'{pygit2.__name__} v{pygit2.__version__} (backed by libgit2 v{pygit2.LIBGIT2_VERSION})'


def get_jinja_info():
    return f'{jinja2.__name__} v{jinja2.__version__}'



import os


class JsPlot:
    def __init__(self, template_filename, **kwargs):
        self.template_filename = template_filename
        self.kwargs = kwargs

    def bootstrap(self, j2_env):
        bootstrapped = j2_env.get_template(self.template_filename).render(
            **self.kwargs,
        )
        return bootstrapped

    @property
    def filename(self):
        # property used for readability and in case in future jinja template name becomes different from
        # rendered filename
        return self.template_filename

    def save(self, path, bootstrapped_js):
        with open(os.path.join(path, self.filename), 'w') as fg:
            fg.write(bootstrapped_js)


class HtmlPage:
    assets_path = None

    def __init__(self, name: str, **kwargs):
        self.name = name
        self.is_active = False
        self.kwargs = kwargs
        self._plots = []
        self._bootstrapped_plots = []

    @classmethod
    def set_assets_path(cls, assets_path):
        cls.assets_path = assets_path

    @property
    def filename(self):
        return self.name.lower() + '.html'

    @property
    def template_name(self):
        return self.name.lower() + '.html'

    def add_plot(self, plot: JsPlot):
        self._plots.append(plot)

    def render(self, j2_env, linked_pages):
        print(f"Rendering '{self.name}'-page")

        # linked_pages contain reference to this page as well
        # the following sets currently rendered page as active to apply appropriate css style in navigation bar
        self.is_active = True
        # load and render template
        template_rendered = j2_env.get_template(self.template_name).render(
            **self.kwargs,
            page_title=self.name,
            pages=linked_pages,
            assets_path=self.assets_path,
        )
        self.is_active = False

        # also bootstrap all page's plots
        for p in self._plots:
            self._bootstrapped_plots.append(p.bootstrap(j2_env))

        return template_rendered.encode('utf-8')

    def save(self, path, rendered):
        # save all plots
        for p, bp in zip(self._plots, self._bootstrapped_plots):
            p.save(path, bp)

        # and the htm itself
        with open(os.path.join(path, self.filename), 'w', encoding='utf-8') as f:
            f.write(rendered.decode('utf-8'))




import os


def split_email_address(email_address):
    parts = email_address.split('@')
    if len(parts) != 2:
        raise ValueError('Not an email passed: %s' % email_address)
    return parts[0], parts[1]


def get_file_extension(filepath: str):
    assert filepath
    filename = os.path.basename(filepath)
    basename_parts = filename.split('.')
    if len(basename_parts) == 1:
        # 'folder/filename'-case
        return filename
    elif len(basename_parts) == 2 and not basename_parts[0]:
        # 'folder/.filename'-case
        return filename
    else:
        # "normal" case
        return basename_parts[-1]



"""
Unused class
"""

import pandas as pd
from datetime import datetime

class GitAuthor(object):
    author_groups = None

    def __init__(self, name):
        self.name = name
        self.group = self.author_groups.get_group(name)

    @property
    def first_commit_date(self):
        timestamp = self.group.min().loc['author_timestamp']
        return datetime.utcfromtimestamp(timestamp)

    @property
    def latest_commit_date(self):
        timestamp = self.group.max().loc['author_timestamp']
        return datetime.utcfromtimestamp(timestamp)

    @property
    def lines_removed(self):
        return self.group['deletions'].sum()

    @property
    def lines_added(self):
        return self.group['insertions'].sum()

    @property
    def active_days_count(self):
        ts = pd.to_datetime(self.group['author_timestamp'], unit='s', utc=True)
        return ts.dt.normalize().unique().shape[0]

    @property
    def contributed_days_count(self):
        if self.first_commit_date != self.latest_commit_date:
            return (self.latest_commit_date - self.first_commit_date).days
        else:
            return 1

    @property
    def commits_count(self):
        return self.group.count().loc['author_timestamp']



import pygit2 as git
import pandas as pd
from typing import List, Generator
from numpy import isnan

from .gitdata import TagsData


class GitTag:

    def __init__(self, tag_df: pd.DataFrame):
        self.df_ref = tag_df

    def __repr__(self):
        return self.name

    @property
    def name(self) -> str:
        tag_name = self.df_ref.tag_name.unique()
        assert len(tag_name) == 1
        tag_name = tag_name[0]
        return tag_name if tag_name is not None else 'unreleased'

    @property
    def contributors(self) -> pd.DataFrame:
        return self.df_ref[['commit_author', 'is_merge']].groupby('commit_author')\
            .count().rename(columns={'is_merge': 'commits_count'}).sort_values(by='commits_count', ascending=False)

    @property
    def created(self):
        """
        This is tagger time, i.e. when tag as created
        :return: timestamp
        """
        ts = self.df_ref['tagger_time'].unique()
        # all tagger_time's for particular tag should be the same
        assert len(ts) == 1
        ts = ts[0]
        return pd.to_datetime(ts, unit='s', utc=True) if ts != -1 else None

    @property
    def initiated(self):
        """
        This is author's time of the very first commit that "belongs" to this tag
        :return: timestamp
        """
        return pd.to_datetime(self.df_ref['commit_time'].min(), unit='s', utc=True)

    @property
    def commits_count(self) -> int:
        """
        How many commits "belong" to this tag
        :return: commits count
        """
        return self.df_ref.shape[0]

    @property
    def tagger(self):
        tagger_name = self.df_ref['tagger_name'].unique()
        assert len(tagger_name) == 1
        tagger_name = tagger_name[0]
        return tagger_name


class GitTags:

    def __init__(self, repo: git.Repository):
        self.tags_data = TagsData(repo).as_dataframe()

    def filter(self, regexp: str) -> List[GitTag]:
        pass

    def all(self) -> 'Generator[GitTag, None]':
        return (self.get(name) for name in self.names)

    def get(self, tag_name: str) -> GitTag:
        predicate = self.tags_data.tag_name.isna() if tag_name is None else self.tags_data.tag_name == tag_name
        return GitTag(self.tags_data[predicate])

    @property
    def names(self):
        return self.tags_data.tag_name.unique()

    @property
    def count(self):
        return len(self.names)




import sys
import os
import argparse


class ReadableDir(argparse.Action):

    def __call__(self, parser, namespace, values, option_string=None):
        prospective_dir = os.path.abspath(os.path.expanduser(values))
        if not os.path.isdir(prospective_dir):
            raise argparse.ArgumentTypeError("Path {0} is not valid.".format(prospective_dir))
        if os.access(prospective_dir, os.R_OK):
            setattr(namespace, self.dest, prospective_dir)
        else:
            raise argparse.ArgumentTypeError("Directory {0} is not readable.".format(prospective_dir))


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(prog='ReportChecker',
                                        description="Checks presence of all expected repostat's report files")
    argparser.add_argument('--is-relocatable', action="store_true", help="Is generated report relocatable")
    argparser.add_argument('--has-index-page', action="store_true", help="Report is generated with index.html")
    argparser.add_argument('report_path', type=str, action=ReadableDir, help="Path to generated report")
    parsed_args = argparser.parse_args(sys.argv[1:])

    expected_files = [
        "about.html",
        "activity.html",
        "activity.js",
        "authors.html",
        "authors.js",
        "files.html",
        "files.js",
        "general.html",
        "tags.html"
    ]

    if parsed_args.has_index_page:
        expected_files.append("index.html")

    # relocatable report has asset files in its dir
    if parsed_args.is_relocatable:
        expected_asset_files = [
            "d3.v3.min.js",
            "gitstats.css",
            os.path.join("images", "arrow-down.gif"),
            os.path.join("images", "arrow-none.gif"),
            os.path.join("images", "arrow-up.gif"),
            "nv.d3.css",
            "nv.d3.min.js",
            "sortable.js",
        ]
        expected_files.extend(os.path.join("assets", asset_file) for asset_file in expected_asset_files)

    existence_check_result = [os.path.exists(os.path.join(parsed_args.report_path, asset_file))
                              for asset_file in expected_files]
    if not all(existence_check_result):
        ifile_not_found = existence_check_result.index(False)
        print(f"File '{expected_files[ifile_not_found]}' has not been found.")
        sys.exit(1)

    print("All good!")



from setuptools import setup
from tools.configuration import Configuration

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt', "r") as req_file:
    requirements = req_file.read().splitlines()

# such name for package is used here because 'repostat' is already occupied by https://pypi.org/project/repostat/
setup(name='repostat-app',
      version=Configuration.get_release_data_info()['develop_version'],
      description='Desktop git repository analyser and report creator.',
      keywords='git analysis statistics vcs python visualization',
      url='https://github.com/vifactor/repostat',
      author='Viktor Kopp',
      author_email='vifactor@gmail.com',
      license='GPLv3',
      long_description=long_description,
      long_description_content_type="text/markdown",
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "Environment :: Console",
          "Intended Audience :: Developers",
          "Intended Audience :: Science/Research",
          "Intended Audience :: Education",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
          "Topic :: Software Development :: Version Control",
          "Topic :: Utilities",
          "Operating System :: OS Independent",
          "Operating System :: POSIX",
          "Operating System :: MacOS :: MacOS X",
          "Operating System :: Microsoft :: Windows",
      ],
      python_requires='>3.5',
      packages=['analysis', 'tools', 'report'],
      package_data={'report': ['templates/*.html',
                               'templates/*.js',
                               'assets/images/*.gif',
                               'assets/*.js',
                               'assets/*.css'],
                    'tools': ['release_data.json']},
      install_requires=requirements,
      entry_points={"console_scripts": ["repostat = analysis.repostat:main"]},
      include_package_data=True,
      zip_safe=False)



import pygit2 as git
import pandas as pd

from tools import get_file_extension
from .gitdata import BlameData, FilesData


class GitRevision:

    def __init__(self, repository: git.Repository, revision: str = 'HEAD'):
        self.blame_data = BlameData(repository, revision)
        self.files_data = FilesData(repository, revision).as_dataframe()

    def _lazy_load_blame_data(self):
        # replaces class with raw DataFrame this class supposes to fetch
        if isinstance(self.blame_data, BlameData):
            self.blame_data = self.blame_data.as_dataframe()

    @property
    def authors_contribution(self):
        self._lazy_load_blame_data()
        return self.blame_data[["committer_name", "lines_count"]]\
            .groupby(by="committer_name")["lines_count"].sum()

    def get_top_files_by_contributors_count(self, top_size=10):
        self._lazy_load_blame_data()
        return self.blame_data[["committer_name", "filepath"]].groupby(["filepath"])\
            .committer_name.nunique().sort_values(ascending=False).head(top_size)

    @property
    def monoauthor_files(self):
        self._lazy_load_blame_data()
        committer_per_file = self.blame_data[["committer_name", "filepath"]].groupby(["filepath"])\
            .committer_name.nunique()
        return committer_per_file[committer_per_file == 1]

    def get_lost_knowledge_percentage(self, knowledge_loss_period_month=6):
        """
        Metrics is introduced in:
        https://www.feststelltaste.de/identifying-lost-knowledge-in-the-linux-kernel-source-code/

        :param knowledge_loss_period_month: months count after which code knowledge is considered to be "lost"
        :return: the ratio of known code to unknown code (= code older than `knowledge_loss_period_month` months)
        """
        self._lazy_load_blame_data()
        months_ago = pd.Timestamp.utcnow() - pd.DateOffset(months=knowledge_loss_period_month)
        df = self.blame_data[["lines_count", "timestamp"]].copy()
        df.timestamp = pd.to_datetime(df.timestamp, unit='s', utc=True)
        df['knowing'] = df.timestamp >= months_ago
        total_lines_count = df.lines_count.sum()
        forgotten_lines_count = df[~df.knowing].lines_count.astype('float').sum()
        return forgotten_lines_count / total_lines_count

    def get_top_knowledge_carriers(self, knowledge_loss_period_month=6):
        """
        Metrics is introduced in:
        https://www.feststelltaste.de/identifying-lost-knowledge-in-the-linux-kernel-source-code/

        :param knowledge_loss_period_month: months count after which code knowledge is considered to be "lost"
        :return: dataframe of contributors with lines count contributed in last `knowledge_loss_period_month`
        """
        self._lazy_load_blame_data()
        months_ago = pd.Timestamp.utcnow() - pd.DateOffset(months=knowledge_loss_period_month)
        df = self.blame_data[["lines_count", "timestamp", 'committer_name']].copy()
        df.timestamp = pd.to_datetime(df.timestamp, unit='s', utc=True)
        df['knowing'] = df.timestamp >= months_ago

        res = df[df.knowing].groupby("committer_name").agg({"lines_count": 'sum'}).reset_index()

        # this step is needed because `committer_name` is categorical and all authors appear to be present
        # in grouped dataframe with zero contribution
        recent_contributors = df[df.knowing].committer_name.unique()
        res = res[res.committer_name.isin(recent_contributors)].sort_values(by="lines_count", ascending=False)\
            .reset_index(drop=True)

        return res

    @property
    def files_count(self):
        return self.files_data["file"].unique().shape[0]

    @property
    def size(self):
        return self.files_data["size_bytes"].sum()

    @property
    def files_extensions_summary(self):
        df = self.files_data[["is_binary", "size_bytes", "lines_count"]].copy()
        df["extension"] = self.files_data['file'].apply(get_file_extension)
        df = df.groupby(by=["is_binary", "extension"]).agg({"size_bytes": ["sum"], "lines_count": ["sum", "count"]})
        df.columns = ["size_bytes", "lines_count", "files_count"]
        df.reset_index()

        return df
